% Calculates the intersection/union for a pair of oriented (with respect to gravity) bounding boxes.
%
% Args:
%   bb1 - struct containing a 3D bounding box.
%   bb2 - struct containing a 3D bounding box.
%
% Returns:
%   score - the intersection / overlap of the two bounding boxes.
%
% Author: Nathan Silberman (silberman@cs.nyu.edu)
function score = get_overlap_bb3d(bb1, bb2)
  % Overlap in Z
  [heightIntersection, heightUnion] = get_height_metrics(bb1, bb2);
  
  % Overlap in the XY plane.
  [rectIntersection, rectUnion] = get_rectangular_metrics(bb1, bb2);
  
  intersection = heightIntersection * rectIntersection;
  union = heightUnion * rectUnion;
  score = intersection / union;
end

function [heightIsect, heightUnion] = get_height_metrics(bb1, bb2)
  maxZ1 = bb1.centroid(3) + bb1.coeffs(3);
  minZ1 = bb1.centroid(3) - bb1.coeffs(3);
  
  maxZ2 = bb2.centroid(3) + bb2.coeffs(3);
  minZ2 = bb2.centroid(3) - bb2.coeffs(3);
  
  maxOfMins = max([minZ1, minZ2]);
  minOfMaxs = min([maxZ1, maxZ2]);
  
  heightIsect = minOfMaxs - maxOfMins;
  heightIsect = max([0, heightIsect]);
  
  heightUnion = max([maxZ1, maxZ2]) - min([minZ1, minZ2]);
  heightUnion = heightUnion - max([0, maxOfMins - minOfMaxs]);
end

function [rectIsect, rectUnion] = get_rectangular_metrics(bb1, bb2)
  ConstsDetection;
  
  [X1, Y1] = get_poly(bb1);
  [X2, Y2] = get_poly(bb2);
  
  maxs = max([[X1; X2], [Y1; Y2]], [], 1);
  mins = min([[X1; X2], [Y1; Y2]], [], 1);
  
  % Create the rectangular image.
  maskDims = ceil((maxs - mins) / consts.rectRes);
  
  % Create the masks for both bounding boxes.
  X1 = (X1 - mins(1)) / consts.rectRes;
  X2 = (X2 - mins(1)) / consts.rectRes;
  Y1 = (Y1 - mins(2)) / consts.rectRes;
  Y2 = (Y2 - mins(2)) / consts.rectRes;
  
  X1 = round(X1);
  Y1 = round(Y1);
  
  mask1 = poly2mask(X1, Y1, maskDims(2), maskDims(1));
  mask2 = poly2mask(X2, Y2, maskDims(2), maskDims(1));
  maskIsect = mask1 & mask2;
  maskUnion = mask1 | mask2;
  
  rectIsect = nnz(maskIsect) * consts.rectRes * consts.rectRes;
  rectUnion = nnz(maskUnion) * consts.rectRes * consts.rectRes;
end

function [X, Y] = get_poly(bb)
  P = zeros(4, 2);
  
  C = bb.centroid(1:2);
  
  P(1,:) = C - bb.basis(1, 1:2) * bb.coeffs(1) + bb.basis(2, 1:2) * bb.coeffs(2);
  P(2,:) = C + bb.basis(1, 1:2) * bb.coeffs(1) + bb.basis(2, 1:2) * bb.coeffs(2);
  P(3,:) = C + bb.basis(1, 1:2) * bb.coeffs(1) - bb.basis(2, 1:2) * bb.coeffs(2);
  P(4,:) = C - bb.basis(1, 1:2) * bb.coeffs(1) - bb.basis(2, 1:2) * bb.coeffs(2);
  
  X = P(:,1);
  Y = P(:,2);
end